import pandas as pd
import os
import re
import xml.etree.cElementTree as ET
from io import StringIO
from PyQt5.QtCore import QObject, pyqtSignal

from cfatools.logreader.dataset import DatasetReader
from cfatools.logreader.pump_calib import PumpCalibReader
import cfatools.processor.iceblock as iceblock_processor
import cfatools.processor.pump as pump_processor
import cfatools.processor.encoder as encoder_processor


class ExploProvider(QObject):

    sig_dataset_loaded = pyqtSignal(DatasetReader, name="dataset_loaded")

    saved_setup_subdir = "saved_setups/"
    saved_setup_ext = ".xml"

    def __init__(self):
        super(ExploProvider, self).__init__()
        self.datasets_root_directory = ""
        self.dataset_readers = {}
        self.current_dataset = None

    def explore_root_directory(self, root_directory: str) -> dict:
        """Get the names of the datasets directories.

        Parameters
        ----------
        root_directory: str
            Full path of the directory containing the datasets directories.

        Returns
        -------
        list
            List of dataset directories name (without full path)

        """
        directories = []

        # Find all directories in datasets root directory (not recursive)
        for element in os.listdir(root_directory):
            if os.path.isdir(root_directory + "/" + element):
                directories.append(element)

        # Keep only datasets directories (ignore others like pump_calibration, conduct_calib, old, etc.)
        regex = re.compile(r"[0-9]{8}_.*")
        dataset_directories = list(filter(regex.search, directories))

        # Sort list in alphabetical order (in this case, by ascending date)
        dataset_directories.sort()

        self.datasets_root_directory = root_directory
        for directory in dataset_directories:
            self.dataset_readers[directory] = DatasetReader(
                base_path=root_directory, dataset=directory
            )

        return self.dataset_readers

    def get_dataset_reader(self, dataset_path: str) -> DatasetReader:
        root_directory = os.path.dirname(dataset_path)
        dataset_name = os.path.basename(dataset_path)
        valid, error_msg = DatasetReader.dataset_name_is_valid(dataset_name)
        if not valid:
            raise ValueError(error_msg)
        self.current_dataset = DatasetReader(
            base_path=root_directory, dataset=dataset_name
        )
        self.sig_dataset_loaded.emit(self.current_dataset)
        return self.current_dataset

    ####################################################################################################################
    # Get instrument's variables and timeseries
    # -----------------------------------------
    # The cfatools' DatasetReader class provides access to raw instrument's log data, which is not always adapted to
    # what the user needs to see.
    # So for common cases, just return raw data from DatasetReader, but if required for a specific instrument log, these
    # data can be modified before being served to the GUI. Just add a get_XXXXX_yyyyy_instrument_variables() and a
    # get_XXXXX_yyyyy_timeseries() function.

    # Default: directly return data from DatasetReader with no post-treatment.
    def get_instrument_variables(
        self, dataset: DatasetReader, instrument_name: str
    ) -> list:
        try:
            custom_func = getattr(
                self, "get_" + instrument_name + "_instrument_variables"
            )
            variables = custom_func(dataset)
        except AttributeError:
            variables = dataset.get_instrument_variables(instrument_name)

        return variables

    def get_timeseries(
        self, dataset: DatasetReader, instrument_name: str, variable: str
    ) -> pd.DataFrame:
        try:
            custom_func = getattr(self, "get_" + instrument_name + "_timeseries")
            timeseries_df = custom_func(dataset, variable)
        except AttributeError:
            timeseries_df = dataset.get_timeseries(instrument_name, variable)

        return timeseries_df

    def get_variable_description(
        self, dataset: DatasetReader, instrument_name: str, variable: str
    ) -> dict:
        try:
            instrument_dict = dataset.instrument_reader.variables_descr[instrument_name]
        except KeyError:
            instrument_dict = {"unit": "", "description": ""}

        try:
            descr_dict = instrument_dict[variable]
        except KeyError:
            descr_dict = {"unit": "", "description": ""}

        try:
            custom_func = getattr(
                self, "get_" + instrument_name + "_variable_description"
            )
            descr_dict = custom_func(instrument_dict, variable)
        except (AttributeError, ValueError):
            pass  # If no custom overriding function exists, or if the overriding function exists, but the requested
            # variable is not recognized.

        return descr_dict

    # ##### ICBKCTRL_instant ##### #
    def get_ICBKCTRL_instant_instrument_variables(self, dataset: DatasetReader) -> list:
        return ["melting"]

    def get_ICBKCTRL_instant_timeseries(
        self, dataset: DatasetReader, variable: str
    ) -> pd.DataFrame:
        raw_df = dataset.get_timeseries("ICBKCTRL_instant")
        timeseries = iceblock_processor.get_melting_timeseries(raw_df)
        return timeseries

    def get_ICBKCTRL_instant_variable_description(
        self, instrument_dict: dict, variable: str
    ):
        return {"unit": "", "description": "Name of the currently-melting ice block."}

    # ##### Peristaltic XXXPUMP_instant ##### #
    def get_PUMP_instant_instrument_variables(
        self, dataset: DatasetReader, pump_id: str
    ) -> list:
        return dataset.get_instrument_variables(pump_id + "_instant") + ["mlmin"]

    def get_PUMP_instant_timeseries(
        self, dataset: DatasetReader, variable: str, pump_id: str
    ) -> pd.DataFrame:
        if variable == "mlmin":
            try:
                rpm_df = dataset.get_timeseries(pump_id + "_instant", "real_rpm")
            except ValueError:
                rpm_df = dataset.get_timeseries(pump_id + "_instant", "rpm")
            calib_reader = PumpCalibReader(
                os.path.join(dataset.base_path, "pump_calibration"), pump_id
            )
            coef_df = calib_reader.get_coef_df()
            timeseries = pump_processor.add_mlmin_to_rpm_history_df(
                rpm_df=rpm_df, coef_df=coef_df, rpm_col_name="value"
            )
            timeseries = timeseries.drop(columns=["value"])
            timeseries.rename(columns={"mlmin": "value"}, inplace=True)
        else:
            timeseries = dataset.get_timeseries(pump_id + "_instant", variable)
        return timeseries

    def get_PUMP_instant_variable_description(
        self, instrument_dict: dict, variable: str
    ):
        if variable == "mlmin":
            return {
                "unit": "ml/min",
                "description": "Pumping flow, based on [ml/min = rpm * slope + intercept] calibration.",
            }
        else:
            raise ValueError("Custom variable [" + variable + "] not recognized.")

    def get_COLPUMP_instant_instrument_variables(self, dataset: DatasetReader) -> list:
        return self.get_PUMP_instant_instrument_variables(dataset, "COLPUMP")

    def get_COLPUMP_instant_timeseries(
        self, dataset: DatasetReader, variable: str
    ) -> pd.DataFrame:
        return self.get_PUMP_instant_timeseries(dataset, variable, "COLPUMP")

    def get_COLPUMP_instant_variable_description(
        self, instrument_dict: dict, variable: str
    ):
        return self.get_PUMP_instant_variable_description(instrument_dict, variable)

    def get_COMPUMP_instant_instrument_variables(self, dataset: DatasetReader) -> list:
        return self.get_PUMP_instant_instrument_variables(dataset, "COMPUMP")

    def get_COMPUMP_instant_timeseries(
        self, dataset: DatasetReader, variable: str
    ) -> pd.DataFrame:
        return self.get_PUMP_instant_timeseries(dataset, variable, "COMPUMP")

    def get_COMPUMP_instant_variable_description(
        self, instrument_dict: dict, variable: str
    ):
        return self.get_PUMP_instant_variable_description(instrument_dict, variable)

    def get_PICPUMP_instant_instrument_variables(self, dataset: DatasetReader) -> list:
        return self.get_PUMP_instant_instrument_variables(dataset, "PICPUMP")

    def get_PICPUMP_instant_timeseries(
        self, dataset: DatasetReader, variable: str
    ) -> pd.DataFrame:
        return self.get_PUMP_instant_timeseries(dataset, variable, "PICPUMP")

    def get_PICPUMP_instant_variable_description(
        self, instrument_dict: dict, variable: str
    ):
        return self.get_PUMP_instant_variable_description(instrument_dict, variable)

    # ##### ENCODER_periodic ##### #
    def get_ENCODER_periodic_instrument_variables(self, dataset: DatasetReader) -> list:
        return [
            "position_mm",
            "speed_mmsec",
            "speed_avg_mmsec",
            "speed_cmmin",
            "speed_avg_cmmin",
        ]

    def get_ENCODER_periodic_timeseries(
        self, dataset: DatasetReader, variable: str
    ) -> pd.DataFrame:
        timeseries = dataset.get_timeseries("ENCODER_periodic")
        timeseries = encoder_processor.add_cmmin_conversion(timeseries)
        timeseries = timeseries[[variable]]
        timeseries.rename(columns={variable: "value"}, inplace=True)
        return timeseries

    def get_ENCODER_periodic_variable_description(
        self, instrument_dict: dict, variable: str
    ):
        if variable == "position_mm":
            return instrument_dict["position"]
        elif variable == "speed_mmsec":
            return instrument_dict["speed"]
        elif variable == "speed_avg_mmsec":
            return instrument_dict["speed_avg"]
        elif variable == "speed_cmmin":
            return {
                "unit": "cm/min",
                "description": instrument_dict["speed"]["description"],
            }
        elif variable == "speed_avg_cmmin":
            return {
                "unit": "cm/min",
                "description": instrument_dict["speed_avg"]["description"],
            }
        else:
            raise ValueError("Custom variable [" + variable + "] not recognized.")

    ####################################################################################################################
    # SAVED_SETUPS

    def save_setup(
        self,
        dataset_reader: DatasetReader,
        setup_name: str,
        variable_df: pd.DataFrame,
        view_range: list,
    ) -> None:
        saved_setup_dir = dataset_reader.dataset_path + self.saved_setup_subdir

        # Build 'saved setup' full file name
        if not os.path.exists(saved_setup_dir):
            os.mkdir(saved_setup_dir)
        filename = saved_setup_dir + setup_name + self.saved_setup_ext

        # Variables table
        variables_str = variable_df.to_csv(sep=";", index=False, mode="w")

        # Create XML file
        root_elmt = ET.Element("save")
        ET.SubElement(root_elmt, "variables").text = variables_str
        view_range_elmt = ET.SubElement(root_elmt, "view_range")
        ET.SubElement(view_range_elmt, "xmin").text = "{:.2f}".format(view_range[0][0])
        ET.SubElement(view_range_elmt, "xmax").text = "{:.2f}".format(view_range[0][1])
        ET.SubElement(view_range_elmt, "ymin").text = "{:.4f}".format(view_range[1][0])
        ET.SubElement(view_range_elmt, "ymax").text = "{:.4f}".format(view_range[1][1])

        tree = ET.ElementTree(root_elmt)
        tree.write(filename)

    def load_setup(self, dataset_reader: DatasetReader, filename: str) -> tuple:
        full_filename = (
            dataset_reader.dataset_path
            + self.saved_setup_subdir
            + filename
            + self.saved_setup_ext
        )

        # Open XML file
        tree = ET.parse(full_filename)
        root = tree.getroot()

        # Variable CSV table as pd.Dataframe
        variables_str = root.find("variables").text
        variable_io = StringIO(variables_str)
        variable_df = pd.read_csv(variable_io, sep=";")

        # View range
        view_range_elmt = root.find("view_range")
        if view_range_elmt is not None:
            view_range_dict = {
                "xmin": float(view_range_elmt.findall("xmin")[0].text),
                "xmax": float(view_range_elmt.findall("xmax")[0].text),
                "ymin": float(view_range_elmt.findall("ymin")[0].text),
                "ymax": float(view_range_elmt.findall("ymax")[0].text),
            }
        else:
            view_range_dict = {}

        return variable_df, view_range_dict

    def setup_filename_is_valid(
        self, dataset_reader: DatasetReader, filename: str
    ) -> tuple:
        """Check if the file name is valid: no special characters, file does not already exists.

        Parameters
        ----------
        filename: str
            filename (without extension) to be tested.

        Returns
        -------
        bool:
            True if the file name is valid, False otherwise
        str:
            The error message explaining why the file name is not valid ; an empty string if file name is valid.
        """
        if not re.match("^[-#&+\w ]*$", filename):
            error_msg = (
                "File name can only contain letters, digits, '-', '_', '#', '&', '+' and whitespace."
                " File extension is automatically set."
            )
            return False, error_msg
        elif filename in self.get_setup_saved_files(dataset_reader):
            error_msg = "File already exists."
            return False, error_msg
        else:
            return True, ""

    def get_setup_saved_files(self, dataset_reader: DatasetReader) -> list:
        """Get a list of the 'setup' file names (without extension) existing in the 'saved_setups' directory."""
        saved_setups_dir = dataset_reader.dataset_path + self.saved_setup_subdir
        if not os.path.exists(saved_setups_dir):
            return []

        filenames = os.listdir(saved_setups_dir)

        files_without_ext = [os.path.splitext(filename)[0] for filename in filenames]
        return files_without_ext
