import datetime
import pyqtgraph as pg
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QColor
from PyQt5.QtCore import *
import pandas as pd
import pytz
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent

import utils
from config import Config
from dataprovider.icbktransitionprovider import IcbktransitionProvider
from uim.explouim import ExploUim
from gui.ui_mainwindow import Ui_MainWindow
from gui.icbktransition_addon_dialog import IcbktransitionAddonDialog


class IcbktransitionUim(QObject):

    # "Variables" tableWidget columns identifiers.
    ID_COL = 0
    NAME_COL = 1
    START_COL = 2
    END_COL = 3
    HEIGHT_COL = 4
    ACTIONS_COL = 5

    PROBES_COLORS = {
        "C1": QColor(228, 26, 28),  # Red #E41A1C
        "C2": QColor(255, 255, 51),  # Yellow #FFFF33
        "C3": QColor(77, 175, 74),  # Green #4DAF4A
        "C4": QColor(152, 78, 163),  # Violet #984EA3
        "C5": QColor(247, 129, 191),  # Pink #F781BF
    }

    PROBES = ["C1", "C2", "C3", "C4", "C5"]

    PEN_TRANSITION_BASE = pg.mkPen(
        style=Qt.SolidLine, width=1, color=QColor(55, 126, 184)
    )  # Blue #377EB8
    PEN_TRANSITION_HOVERED = pg.mkPen(
        style=Qt.DotLine, width=3, color=QColor(55, 126, 184)
    )  # Blue #377EB8
    PEN_TRANSITION_SELECTED = pg.mkPen(
        style=Qt.DotLine, width=3, color=QColor(255, 127, 0)
    )  # Orange #FF7F00

    sig_conduct_reftime_df_updated = pyqtSignal()
    sig_hovered_transition_updated = pyqtSignal(int)
    sig_selected_transition_update = pyqtSignal(int)

    def __init__(
        self,
        icbktransition_prvd: IcbktransitionProvider,
        explo_uim: ExploUim,
        main_ui: Ui_MainWindow,
        config: Config,
        icbktransition_dialog: IcbktransitionAddonDialog,
    ):
        super(IcbktransitionUim, self).__init__()
        self.main_ui = main_ui
        self.prvd = icbktransition_prvd
        self.explo_uim = explo_uim
        self.config = config
        self.icbktransition_dialog = icbktransition_dialog
        self.icbktransition_ui = self.icbktransition_dialog.ui

        self.prvd.sig_transition_dataset_loaded.connect(
            lambda: self.main_ui.explo_pushbutton_icbktransition.setEnabled(True)
        )

        self.__init_plot__()
        self.main_ui.explo_pushbutton_icbktransition.clicked.connect(
            self.icbktransition_dialog.show
        )
        self.main_ui.explo_pushbutton_icbktransition.clicked.connect(
            self.__initialize_all__
        )
        self.icbktransition_ui.timeedit_reftime.dateTimeChanged.connect(
            self.__update_conduct_reftime_df__
        )
        self.icbktransition_ui.checkbox_reftime_disable.clicked.connect(
            self.__update_conduct_reftime_df__
        )

        self.__initialize_transition_table__()
        self.icbktransition_ui.toolbutton_reload_icbkctrl.clicked.connect(
            self.__reload_from_icbkctrl__
        )
        self.icbktransition_ui.toolbutton_reload_iceblocks.clicked.connect(
            self.__reload_from_iceblocks__
        )
        self.icbktransition_ui.toolbutton_save_iceblocks.clicked.connect(
            self.prvd.save_iceblock_df_as_csv
        )
        self.icbktransition_dialog.sig_escape_pressed.connect(
            self.__unselect_transition__
        )

        self._hovered_transition = None
        self.sig_hovered_transition_updated.connect(self.__paint_transitions_in_table__)
        self.sig_hovered_transition_updated.connect(self.__paint_transitions_in_plot__)

        self._selected_transition = None
        self.sig_selected_transition_update.connect(self.__paint_transitions_in_table__)
        self.sig_selected_transition_update.connect(self.__paint_transitions_in_plot__)

        self.prvd.sig_transition_dataset_loaded.connect(self.__initialize_all__)
        self.prvd.sig_transitions_df_updated.connect(
            self.__reload_transitions_in_plot__
        )
        self.prvd.sig_iceblock_df_updated.connect(
            self.__load_iceblocks_in_transition_table__
        )

        self._conduct_reftime_df = pd.DataFrame()
        self.sig_conduct_reftime_df_updated.connect(self.__update_conduct_curves__)

    @property
    def conduct_reftime_df(self) -> pd.DataFrame:
        return self._conduct_reftime_df

    @conduct_reftime_df.setter
    def conduct_reftime_df(self, conduct_reftime_df):
        self._conduct_reftime_df = conduct_reftime_df
        self.sig_conduct_reftime_df_updated.emit()

    @property
    def hovered_transition(self) -> int:
        return self._hovered_transition

    @hovered_transition.setter
    def hovered_transition(self, hovered_transition):
        if hovered_transition != self._hovered_transition:
            self._hovered_transition = hovered_transition
            self.sig_hovered_transition_updated.emit(hovered_transition)

    @property
    def selected_transition(self) -> int:
        return self._selected_transition

    @selected_transition.setter
    def selected_transition(self, selected_transition):
        if selected_transition != self._selected_transition:
            self._selected_transition = selected_transition
            self.sig_selected_transition_update.emit(selected_transition)

    def __initialize_all__(self):
        self.__init_reftime__()
        self.__update_conduct_reftime_df__()
        self.__load_iceblocks_in_transition_table__()
        self.__reload_transitions_in_plot__()

    ####################################################################################################################
    # "Transition" table

    def __initialize_transition_table__(self):
        # Set column widths
        self.icbktransition_ui.tablewidget_icbk_list.setColumnWidth(self.ID_COL, 30)
        self.icbktransition_ui.tablewidget_icbk_list.setColumnWidth(self.NAME_COL, 130)
        self.icbktransition_ui.tablewidget_icbk_list.setColumnWidth(self.START_COL, 90)
        self.icbktransition_ui.tablewidget_icbk_list.setColumnWidth(self.END_COL, 90)
        self.icbktransition_ui.tablewidget_icbk_list.setColumnWidth(
            self.HEIGHT_COL, 100
        )
        self.icbktransition_ui.tablewidget_icbk_list.setColumnWidth(
            self.ACTIONS_COL, 120
        )

    def __reload_from_icbkctrl__(self):
        iceblock_df = self.prvd.build_iceblock_df(
            retrieve_iceblocks_from_processed_csv=False
        )

    def __reload_from_iceblocks__(self):
        iceblock_df = self.prvd.build_iceblock_df(
            retrieve_iceblocks_from_processed_csv=True
        )

    ####################################################################
    # Transitions

    def __select_transition__(self):
        if self.hovered_transition is None:
            return
        self.selected_transition = self.hovered_transition

    def __unselect_transition__(self):
        self.selected_transition = None

    def __move_transition__(self, timestamp_sec: int):
        self.prvd.move_transition(self.selected_transition, timestamp_sec)
        self.originaltransition_vlines[self.selected_transition].setPos(timestamp_sec)
        self.selected_transition = None

    def __mark_hovered_transition__(self, timestamp_sec: int):
        """Determine which transition is hovered on the plot (if any) and repaint plot and table"""
        df = self.prvd.transitions_df.copy()
        df["interval"] = abs(df["timestamp"] - timestamp_sec)
        df = df[df["interval"] < (5 * 60)]

        if len(df.index) == 0:
            self.hovered_transition = None
        else:
            self.hovered_transition = df.loc[[df["interval"].idxmin()]].index[0]

        # self.__paint_transitions_in_plot__()

    ####################################################################################################################
    # Table modifications

    def __load_iceblocks_in_transition_table__(self, iceblock_df: pd.DataFrame = None):
        table = self.icbktransition_ui.tablewidget_icbk_list

        if iceblock_df is None:
            iceblock_df = self.prvd.build_iceblock_df()

        # Clear table
        table.setRowCount(0)

        # Load iceblock_df in the table
        for row_id, iceblock in iceblock_df.iterrows():
            table.insertRow(row_id)

            # ID
            id_item = QTableWidgetItem()
            id_item.setFlags(Qt.ItemIsEnabled)  # Read only
            id_item.setText(str(iceblock["id"]))
            table.setItem(row_id, self.ID_COL, id_item)

            # NAME
            name_item = QTableWidgetItem()
            name_item.setFlags(Qt.ItemIsEnabled)  # Read only
            name_item.setText(str(iceblock["name"]))
            table.setItem(row_id, self.NAME_COL, name_item)

            # START
            start_item = QTableWidgetItem()
            start_item.setFlags(Qt.ItemIsEnabled)  # Read only
            start_item.setText(iceblock["datetime_start"].strftime("%H:%M:%S"))
            table.setItem(row_id, self.START_COL, start_item)

            # END
            end_item = QTableWidgetItem()
            end_item.setFlags(Qt.ItemIsEnabled)  # Read only
            if iceblock["datetime_end"] is not pd.NaT:
                end_item.setText(iceblock["datetime_end"].strftime("%H:%M:%S"))
            table.setItem(row_id, self.END_COL, end_item)

            # HEIGHT
            height_item = QTableWidgetItem()
            height_item.setFlags(Qt.ItemIsEnabled)  # Read only
            height_item.setText(str(iceblock["initial_height"]))
            table.setItem(row_id, self.HEIGHT_COL, height_item)

            # ACTIONS
            actions_item = QPushButton("Delete")
            table.setCellWidget(row_id, self.ACTIONS_COL, actions_item)
            table.cellWidget(row_id, self.ACTIONS_COL).clicked.connect(
                lambda checked, row_id=row_id: self.prvd.delete_transition(
                    transition_id=row_id
                )
            )
        table.cellClicked.connect(self.__cell_clicked__)

    def __cell_clicked__(self, row_id: int, column_id: int):
        if column_id == self.START_COL:
            transition_id = row_id
        elif column_id == self.END_COL:
            transition_id = row_id + 1
        else:
            return
        self.__zoom_transition__(transition_id=transition_id)

    def __paint_transitions_in_table__(self):
        # First, reset the display
        table = self.icbktransition_ui.tablewidget_icbk_list
        for row_id in range(table.rowCount()):
            table.item(row_id, self.START_COL).setBackground(QColor("white"))
            table.item(row_id, self.END_COL).setBackground(QColor("white"))

        if self.hovered_transition is not None:
            self.__paint_single_transition_in_table__(self.hovered_transition, "blue")
        if self.selected_transition is not None:
            self.__paint_single_transition_in_table__(
                self.selected_transition, "orange"
            )

    def __paint_single_transition_in_table__(self, transition_id: int, color: str):
        table = self.icbktransition_ui.tablewidget_icbk_list
        if transition_id < table.rowCount():
            table.item(transition_id, self.START_COL).setBackground(QColor(color))
            table.scrollToItem(table.item(transition_id, self.ID_COL))
        if transition_id > 0:
            table.item(transition_id - 1, self.END_COL).setBackground(QColor(color))

    ####################################################################################################################
    # Plot modifications

    def __init_plot__(self):
        self.plot_item = pg.PlotItem(
            axisItems={"bottom": utils.TimeAxisItem(orientation="bottom")}
        )
        self.icbktransition_dialog.ui.graphicsview.setCentralItem(self.plot_item)
        self.plot_item.sigRangeChanged.connect(self.__autorescale_picarro_curve__)

        # Conduct curves
        self.conduct_curves = dict()
        for probe in self.PROBES:
            self.conduct_curves[probe] = pg.PlotCurveItem()
            self.plot_item.addItem(self.conduct_curves[probe])

        # Picarro curve
        self.picarro_curve = pg.PlotCurveItem()
        self.plot_item.addItem(self.picarro_curve)

        # Vertical lines for the original iceblocks transition
        self.originaltransition_vlines = []
        self.originaltransition_texts = []

        # Vertical lines for the original iceblocks transition
        self.newtransition_vlines = []
        self.newtransition_texts = []

        # Vertical line following the cursor
        self.cursor_vline = pg.InfiniteLine(
            angle=90, movable=False, pen=pg.mkPen(style=Qt.DotLine)
        )
        self.plot_item.addItem(self.cursor_vline, ignoreBounds=True)
        self.conduct_curves["C1"].scene().sigMouseMoved.connect(
            lambda event: self.__mouse_moved__(event)
        )
        self.conduct_curves["C1"].scene().sigMouseClicked.connect(
            lambda event: self.__mouse_clicked__(event)
        )

    def __autorescale_picarro_curve__(self, viewbox: pg.ViewBox, range: list):
        range_x = range[0]
        range_y = range[1]

        rescaled_picarro_df = self.prvd.picarro_df.copy()
        x_min = pd.Timestamp(range_x[0] * 1000000000, tz="UTC")
        x_max = pd.Timestamp(range_x[1] * 1000000000, tz="UTC")
        top = (
            range_y[1] - (range_y[1] - range_y[0]) * 0.05
        )  # Make sure the picarro curve does not "touch" the top
        bottom = (
            range_y[0] + (range_y[1] - range_y[0]) * 0.05
        )  # of the viewbox to avoid infinite auto-rescale.
        picarro_visible_df = self.prvd.picarro_df[x_min:x_max]
        picarro_min = picarro_visible_df["Delta_D_H"].min()
        picarro_max = picarro_visible_df["Delta_D_H"].max()

        rescaled_picarro_df["rescaled_Delta_D_H"] = (top - bottom) * (
            rescaled_picarro_df["Delta_D_H"] - picarro_min
        ) / (picarro_max - picarro_min) + bottom

        x_values = utils.pd_time_to_epoch_ms(rescaled_picarro_df.index)
        x_values.append(x_values[-1] + 1)
        y_values = list(rescaled_picarro_df["rescaled_Delta_D_H"])
        self.picarro_curve.setData(
            x=x_values,
            y=y_values,
            pen=QColor(166, 86, 40),  # Brown #A65628
            stepMode=True,
        )

    def __mouse_moved__(self, pos: QPointF) -> None:
        """Function triggered when the user's mouse cursor hovers over the plot. Display a vertical line where the
        cursor is.

        Parameters
        ----------
        pos: PyQt5.QtCore.QPointF
        """
        mouse_point = self.conduct_curves["C1"].getViewBox().mapSceneToView(pos)

        self.cursor_vline.setPos(mouse_point.x())
        self.__update_timeedit_cursor__(mouse_point.x())
        self.__update_distance_to_transition(mouse_point.x())

        if self.selected_transition is None:
            self.__mark_hovered_transition__(mouse_point.x())

    def __mouse_clicked__(self, event: MouseClickEvent) -> None:
        """Function triggered when the user clicks on the plot. Display a vertical line under the mouse click and call
        function ``__update_xy_original__``

        Parameters
        ----------
        event: pyqtgraph.MouseClickEvent
        """
        pos = event.scenePos()
        mouse_point = self.conduct_curves["C1"].getViewBox().mapSceneToView(pos)
        if self.selected_transition is None:
            self.__select_transition__()
        else:
            self.__move_transition__(mouse_point.x())

    def __zoom_transition__(self, transition_id: int):
        if transition_id >= len(
            self.prvd.transitions_df.index
        ):  # In case last end_datetime is NaT
            return
        transition = self.prvd.transitions_df.iloc[transition_id]
        xmin = transition["timestamp"] - 5 * 60
        xmax = transition["timestamp"] + 5 * 60

        visible_df = self.conduct_reftime_df[
            pd.Timestamp(xmin * 10**9, tz="UTC") : pd.Timestamp(
                xmax * 10**9, tz="UTC"
            )
        ].drop("C1", axis=1)
        conduct_min = visible_df.min().min()
        conduct_max = visible_df.max().max()
        ymin = conduct_min - (conduct_max - conduct_min) * 0.1
        ymax = conduct_max + (conduct_max - conduct_min) * 0.1
        self.plot_item.getViewBox().setRange(
            xRange=(xmin, xmax), yRange=(ymin, ymax), padding=0
        )

        # Update hoovered transition
        self.hovered_transition = transition_id

    def __reload_transitions_in_plot__(self):
        """Delete any existing transition vline and re-create then, based on Provider's transitions_df."""
        for transition_vline in self.originaltransition_vlines:
            self.plot_item.removeItem(transition_vline)
        self.originaltransition_vlines = []

        transitions_df = self.prvd.transitions_df
        transitions_df["timestamp"] = utils.pd_time_to_epoch_ms(
            transitions_df["datetime"]
        )
        for index, transition in transitions_df.iterrows():
            transition_line = pg.InfiniteLine(
                angle=90, movable=False, pen=self.PEN_TRANSITION_BASE
            )
            transition_line.setPos(transition["timestamp"])
            self.originaltransition_vlines.append(transition_line)
            self.plot_item.addItem(transition_line, ignoreBounds=True)

    def __paint_transitions_in_plot__(self):
        # First, reset the display
        for transition_vline in self.originaltransition_vlines:
            transition_vline.setPen(self.PEN_TRANSITION_BASE)

        if self.hovered_transition is not None:
            self.originaltransition_vlines[self.hovered_transition].setPen(
                self.PEN_TRANSITION_HOVERED
            )

        if self.selected_transition is not None:
            self.originaltransition_vlines[self.selected_transition].setPen(
                self.PEN_TRANSITION_SELECTED
            )

    def __update_conduct_curves__(self):
        # Set data to step curve
        x_values = utils.pd_time_to_epoch_ms(self.conduct_reftime_df.index)
        x_values.append(x_values[-1] + 1)
        for probe in self.PROBES:
            y_values = list(self.conduct_reftime_df[probe])
            self.conduct_curves[probe].setData(
                x=x_values, y=y_values, pen=self.PROBES_COLORS[probe], stepMode=True
            )

    ####################################################################################################################
    # Reftime: use a common reference point for all conducti probes

    def __init_reftime__(self):
        min_reftime = min(self.prvd.conduct_df.index)
        self.icbktransition_ui.timeedit_reftime.blockSignals(True)
        self.icbktransition_ui.timeedit_reftime.setMinimumDateTime(min_reftime)
        self.icbktransition_ui.timeedit_reftime.setMaximumDateTime(
            max(self.prvd.conduct_df.index)
        )
        self.icbktransition_ui.timeedit_reftime.setDateTime(
            min_reftime + datetime.timedelta(minutes=20)
        )
        self.icbktransition_ui.timeedit_reftime.blockSignals(False)

    def __update_conduct_reftime_df__(self):
        if self.icbktransition_ui.checkbox_reftime_disable.isChecked():
            self.conduct_reftime_df = self.prvd.conduct_df
        else:
            timezone = pytz.timezone("UTC")
            reftime = self.icbktransition_ui.timeedit_reftime.dateTime().toPyDateTime()
            reftime = timezone.localize(reftime)
            self.conduct_reftime_df = self.prvd.conduct_df.apply(
                self.__shift_to_reftime__, args=(reftime,)
            )

    def __shift_to_reftime__(self, series: pd.Series, reftime) -> pd.Series:
        value_at_reftime = series.iloc[series.index.get_loc(reftime, method="nearest")]
        series = series - value_at_reftime
        return series

    ####################################################################################################################
    # Tiemedit cursor

    def __update_timeedit_cursor__(self, timestamp_sec: int):
        pos_x = QDateTime()
        pos_x.setMSecsSinceEpoch(timestamp_sec * 1000)
        pos_x = pos_x.toUTC()
        self.icbktransition_ui.timeedit_cursor.setDateTime(pos_x)

    def __update_distance_to_transition(self, timestamp_sec: int):
        if self.selected_transition is not None:
            transition_id = self.selected_transition
            transition_type = "selected"
        elif self.hovered_transition is not None:
            transition_id = self.hovered_transition
            transition_type = "hovered"
        else:
            self.icbktransition_ui.label_distancetotransition.setText("")
            return
        transition_info = self.prvd.transitions_df.iloc[transition_id]

        # Distance in seconds
        distance_sec = timestamp_sec - transition_info["timestamp"]

        # Distance in mm
        position_transition = self.prvd.get_encoder_height_at(
            transition_info["datetime"]
        )
        position_cursor = self.prvd.get_encoder_height_at(
            pd.Timestamp(timestamp_sec * 10**9, tz="UTC")
        )
        distance_mm = position_cursor - position_transition

        self.icbktransition_ui.label_distancetotransition.setText(
            "Distance to "
            + transition_type
            + ": "
            + "{:.1f}".format(distance_sec)
            + "s & "
            + "{:.1f}".format(distance_mm)
            + "mm"
        )
