import os
import datetime
import pytz
from math import atan, degrees
import pyqtgraph as pg
from PyQt5.QtCore import *
import numpy as np
import pandas as pd

from cfatools.logreader.pump_calib import PumpCalibReader
import cfatools.processor.pump as pump_processor

import utils
from config import Config
from gui.ui_mainwindow import Ui_MainWindow


class PumpCalibUim:
    def __init__(self, main_ui: Ui_MainWindow, config: Config):
        self.main_ui = main_ui
        self.config = config

        self.pump_reader = None
        self.main_ui.pump_pushbutton_load.clicked.connect(self.load_data)

        self.runs_df = pd.DataFrame()
        self.coef_df = pd.DataFrame()

        now = datetime.datetime.now(tz=datetime.timezone.utc)
        self.main_ui.pump_dateedit_from.setDate(now - datetime.timedelta(days=30))
        self.main_ui.pump_dateedit_to.setDate(now)
        self.main_ui.pump_dateedit_from.dateChanged.connect(self.__update_runs_plot__)
        self.main_ui.pump_dateedit_to.dateChanged.connect(self.__update_runs_plot__)
        self.main_ui.pump_dateedit_from.dateChanged.connect(self.__update_evol_plot__)
        self.main_ui.pump_dateedit_to.dateChanged.connect(self.__update_evol_plot__)

    def load_data(self):
        # Get the directory containing the pump calibrations' files.
        directory = utils.get_data_path(self.config, "pump_calibs", "pump calibrations")
        if directory == "":
            return

        # Get pump reader
        base_path, pump_id = os.path.split(directory)
        try:
            self.pump_reader = PumpCalibReader(base_path, pump_id)
            self.runs_df = self.pump_reader.get_runs_df()
            self.coef_df = self.pump_reader.get_coef_df()
            self.__update_runs_plot__()
            turns_history_df = pump_processor.get_turns_history_df(
                rpm_history_df=self.pump_reader.get_state_changes_df("rpm"),
                turning_history_df=self.pump_reader.get_state_changes_df("turning"),
            )
            self.evol_df = pump_processor.get_mlmin_to_rpm_evol_df(
                turns_history_df=turns_history_df, runs_df=self.runs_df
            )
        except Exception as e:
            utils.show_popup(str(e), "Could not load pump's calibration")
            return

        self.__update_evol_plot__()
        self.main_ui.pump_lineedit_pumpid.setText(pump_id)

    ####################################################################
    # Plot

    def __update_runs_plot__(self) -> None:
        if len(self.runs_df.index) == 0:  # No pump selected yet
            return

        # Create plot item
        plot_item = pg.PlotItem(labels={"left": "ml/min", "bottom": "RPM"})
        self.main_ui.pump_graphicsview_runs.setCentralItem(plot_item)

        # Create legend
        legend = pg.LegendItem(offset=(57, 5))
        legend.setParentItem(plot_item.graphicsItem())

        # Keep only data within defined time range
        first_date = pytz.utc.localize(
            self.main_ui.pump_dateedit_from.dateTime().toPyDateTime()
        )
        last_date = pytz.utc.localize(
            self.main_ui.pump_dateedit_to.dateTime().toPyDateTime()
        )
        runs_df = self.runs_df[
            (self.runs_df["start_time"] > first_date)
            & (self.runs_df["end_time"] < last_date)
        ]
        coef_df = self.coef_df[
            (self.coef_df["first_run_datetime"] > first_date)
            & (self.coef_df["first_run_datetime"] < last_date)
        ]

        # Create the color scale
        color_map = pg.ColorMap(
            pos=runs_df["calib_id_int"].quantile(np.linspace(0, 1, 8).tolist()),
            color=[
                [228, 26, 28],
                [55, 126, 184],
                [77, 175, 74],
                [152, 78, 163],
                [255, 127, 0],
                [255, 255, 51],
                [166, 86, 40],
                [247, 129, 191],
            ],
        )

        # Add linear regression line
        for row_id, row in coef_df.iterrows():
            reg_line = pg.InfiniteLine(
                pos=QPointF(0, row["intercept"]),
                angle=degrees(atan(row["slope"])),
                pen=pg.mkPen(color=color_map.mapToQColor(row["calib_id_int"])),
                name=row["calib_id_str"],
            )
            plot_item.addItem(reg_line)

        # Add runs points
        for calib_id in runs_df["calib_id_int"].unique():
            run_df = runs_df[runs_df["calib_id_int"] == calib_id]
            run_points = pg.ScatterPlotItem(
                run_df["rpm"],
                run_df["flow"],
                size=6,
                brush=color_map.mapToQColor(calib_id),
                name=run_df["calib_id_str"].unique()[0],
            )
            legend.addItem(run_points, run_df["calib_id_str"].unique()[0])
            plot_item.addItem(run_points)

    def __update_evol_plot__(self) -> None:
        if len(self.runs_df.index) == 0:  # No pump selected yet
            return

        # Create plot item
        plot_item = pg.PlotItem(labels={"left": "ml/min", "bottom": "Total rounds"})
        self.main_ui.pump_graphicsview_repro.setCentralItem(plot_item)

        # Create legend
        legend = pg.LegendItem(offset=(57, 5))
        legend.setParentItem(plot_item.graphicsItem())

        # Keep only data within defined time range
        first_date = pytz.utc.localize(
            self.main_ui.pump_dateedit_from.dateTime().toPyDateTime()
        )
        last_date = pytz.utc.localize(
            self.main_ui.pump_dateedit_to.dateTime().toPyDateTime()
        )
        evol_df = self.evol_df[
            (self.evol_df["start_time"] > first_date)
            & (self.evol_df["start_time"] < last_date)
        ]

        # Create the color scale
        color_map = pg.ColorMap(
            pos=[evol_df["rpm"].min(), evol_df["rpm"].max()],
            color=[[255, 0, 255], [255, 0, 0]],
        )

        for rpm in evol_df["rpm"].unique():
            rpm_df = evol_df[evol_df["rpm"] == rpm]
            rpm_points = pg.ScatterPlotItem(
                rpm_df["total_turns"],
                rpm_df["flow"],
                size=6,
                brush=color_map.mapToQColor(rpm),
                name=rpm,
            )
            legend.addItem(rpm_points, rpm)
            plot_item.addItem(rpm_points)
